/* copyright (c)2012..2013 Oskar Schirmer <oskar@scara.com> */
#include "chanend.h"
#include "dispatch.h"
#include "ethernet.h"
#include "execuser.h"
#include "init.h"
#include "print.h"
#include "reception.h"
#include "resource.h"
#include "uart.h"

static void reception(void)
{
  resource c;
  unsigned count = 0;
  unsigned await = 0;
  struct {
    unsigned who;
    unsigned chn;
  } rcpt[128];
  struct {
    unsigned who;
    unsigned chn;
  } requ[128];
  c = chanend_book();
  if (c == chanend_cr) do {
    unsigned char req;
    unsigned who, chn;
    int i;
    chn = chanend_listen(c);
    req = chanend_read_byte(c);
    who = chanend_read_word(c);
    i = count;
    while (--i >= 0)
      if (rcpt[i].who == who)
        break;
    switch (req) {
      case cr_cmd_welcome:
        if ((i < 0) && (count < 128))
          i = count++;
        if (i >= 0) {
          rcpt[i].who = who;
          rcpt[i].chn = chn;
        }
        chanend_hangup(c);
        i = 0;
        while (i < await)
          if (requ[i].who == who) {
            chanend_destination(c, requ[i].chn);
            chanend_write_word(c, chn);
            chanend_all_sent(c);
            requ[i] = requ[--await];
          } else {
            i += 1;
          }
        break;
      case cr_cmd_request:
        if (i >= 0) {
          chanend_write_word(c, rcpt[i].chn);
          chanend_hangup(c);
        } else {
          if (await < 128) {
            i = await++;
            requ[i].who = who;
            requ[i].chn = chn;
          }
          chanend_not_all_sent(c);
          chanend_all_received(c);
        }
        break;
      default:
        chanend_hangup(c);
        break;
    }
  } while (1);
  chanend_free(c);
}

#define sm_cmd_mine 1
#define sm_cmd_drop 2
#define sm_cmd_task 3
#define sm_cmd_done 4

static unsigned schedule_threadsfree()
{
  unsigned r, i;
  resource tmp[8];
  i = 0;
  while (tmp[i] = thread_book()) {
    i += 1;
  }
  r = i;
  while (i > 0) {
    thread_free(tmp[--i]);
  }
  return r;
}

#define PAGESHIFT 10
#define PAGESIZE (1 << PAGESHIFT)
#define NUMPAGES (0x10000 / PAGESIZE)
static unsigned bss_dummy[1];
static void inline schedule(const resource log)
{
  unsigned i, e;
  resource c, disp;
  signed char mused[NUMPAGES];
  void *membase;
  membase = (void *)ps_get(0);
  for (i = 0; i < NUMPAGES; i++) {
    mused[i] = -1;
  }
  i = schedule_threadsfree();
  c = chanend_book();
  disp = chanend_cr_request(c, cr_init_dispatch);
  chanend_connect(c, disp);
  chanend_write_byte(c, dp_cmd_core);
  chanend_write_word(c, i);
  e = ((unsigned)&mused[-64] - (unsigned)membase) >> PAGESHIFT;
  i = (((unsigned)&bss_dummy[0] - (unsigned)membase) >> PAGESHIFT) + 1;
  chanend_write_word(c, (e - i) << PAGESHIFT);
  chanend_hangup(c);
  while (i < e) {
    mused[i++] = 0;
  }
#ifdef DEBUG_SCHED
  printstr(log, "SCHED= ");
  printhex(log, c, 8);
  println(log);
#endif
  do {
    unsigned char req, ctrl;
    unsigned val, en, cp, dp;
    unsigned char *tsk = 0;
    resource r;
    chanend_listen(c);
    req = chanend_read_byte(c);
#ifdef DEBUG_SCHED
    printstr(log, "SCHED: ");
    printhex(log, req, 2);
    printchar(log, ' ');
    printhex(log, val, 8);
    println(log);
#endif
    switch (req) {
      case sm_cmd_task:
        val = (chanend_read_word(c) + PAGESIZE - 1) >> PAGESHIFT;
        i = NUMPAGES;
        e = val;
        do {
          if (!mused[--i]) {
            if (--e == 0) {
              tsk = (unsigned)membase + (i << PAGESHIFT);
              e = val;
              val <<= PAGESHIFT;
#ifdef DEBUG_SCHED
              printstr(log, "SCHED M: ");
              printhex(log, tsk, 8);
              printchar(log, ' ');
              printhex(log, val, 8);
              println(log);
#endif
              en = chanend_read_word(c);
              cp = chanend_read_word(c);
              dp = chanend_read_word(c);
              chanend_write_word(c, val);
#ifdef DEBUG_SCHED
              printstr(log, "SCHED D: ");
              printhex(log, en, 8);
              printchar(log, ' ');
              printhex(log, cp, 8);
              printchar(log, ' ');
              printhex(log, dp, 8);
              println(log);
#endif
              chanend_read_lsb(c, tsk, val);
              r = thread_start(en + (unsigned)tsk, dp + (unsigned)tsk,
                  tsk + val - 4, cp + (unsigned)tsk, disp);
              if (r) {
                ctrl = resource_index(r);
                do {
                  mused[i + --e] = ctrl;
                } while (e > 0);
              }
              i = r;
              break;
            }
          } else {
            e = val;
          }
        } while (i > 0);
        chanend_write_word(c, i);
        chanend_hangup(c);
#ifdef DEBUG_SCHED
        printstr(log, "SCHED I: ");
        printhex(log, i, 8);
        println(log);
#endif
        break;
      case sm_cmd_done:
        i = chanend_read_word(c);
        ctrl = resource_index(i);
        e = 0;
#ifdef DEBUG_SCHED
        printstr(log, "SCHED-");
        printhex(log, ctrl, 2);
        printchar(log, ':');
        println(log);
#endif
        for (i = 0; i < NUMPAGES; i++) {
#ifdef DEBUG_SCHED
          printchar(log, mused[i] < 0 ? 'X' : mused[i] ? '0' + mused[i] : '-');
#endif
          if (mused[i] == ctrl) {
            mused[i] = 0;
            e += 1;
          } else if (!mused[i]) {
            e += 1;
          }
        }
#ifdef DEBUG_SCHED
        println(log);
#endif
        i = schedule_threadsfree();
        chanend_write_word(c, i);
        chanend_write_word(c, e << PAGESHIFT);
        chanend_hangup(c);
        break;
      default:
        chanend_hangup(c);
        break;
    }
  } while (1);
}

static void dispatch(void)
{
  resource c, d;
#ifdef DEBUG_DISPATCH
  resource log;
#endif
  struct {
    resource sched;
    unsigned thr, mem;
  } core[4];
  unsigned cores = 0;
  c = chanend_book();
  chanend_cr_welcome(c, cr_init_dispatch);
  d = chanend_book();
#ifdef DEBUG_DISPATCH
  log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
#endif
  do {
    unsigned char req;
    unsigned i, k, v;
    resource r;
    unsigned data[7];
    v = chanend_listen(c);
    req = chanend_read_byte(c);
    switch (req) {
      case dp_cmd_core:
        for (i = 0; i < cores; i++) {
          if (core[i].sched == v)
            break;
        }
        if (i >= cores) {
          core[i].sched = v;
          cores += 1;
        }
        core[i].thr = chanend_read_word(c);
        core[i].mem = chanend_read_word(c);
        chanend_hangup(c);
#ifdef DEBUG_DISPATCH
        printstr(log, "*** DP STAT ");
        printhex(log, core[i].thr, 2);
        printchar(log, ' ');
        printhex(log, core[i].mem, 8);
        println(log);
#endif
        break;
      case dp_cmd_task:
        r = 0;
        i = chanend_read_lsb(c, &data[0], 28);
        if ((i == 28) && (data[0] == 0xbece)) {
          k = 0;
          v = 0;
          if (data[1] & (dp_task_flag_samecore | dp_task_flag_wantclose)) {
            for (i = 0; i < cores; i++) {
              if (resource_compatible(core[i].sched, data[1], 1, 1, 0, 0)) {
                if (core[i].thr && (core[i].mem > v)) {
                  k = i;
                  v = core[i].mem;
                }
                break;
              }
            }
          }
          if (!(data[1] & dp_task_flag_samecore)) {
            if (data[1] & dp_task_flag_wantclose) {
              for (i = 0; i < cores; i++) {
                if (resource_compatible(core[i].sched, data[1], 1, 0, 0, 0)) {
                  if (core[i].thr && (core[i].mem > v)) {
                    k = i;
                    v = core[i].mem;
                  }
                }
              }
            }
            for (i = 0; i < cores; i++) {
              if (core[i].thr && (core[i].mem > v)) {
                k = i;
                v = core[i].mem;
              }
            }
          }
#ifdef DEBUG_DISPATCH
          printstr(log, "*** DP TASK ");
          printhex(log, core[k].sched, 8);
          printchar(log, ' ');
          printhex(log, v, 8);
          printchar(log, ' ');
          printhex(log, data[2], 8);
          printchar(log, ' ');
          printhex(log, data[3], 8);
          println(log);
#endif
          if (v && (core[k].mem >= (data[2] + data[3] - 16))) {
            chanend_connect(d, core[k].sched);
            chanend_write_byte(d, sm_cmd_task);
            chanend_write_word(d, data[2] + data[3] - 16);
            chanend_write_word(d, data[4] - 12);
            chanend_write_word(d, data[5] - 8);
            chanend_write_word(d, data[6] - 4);
            v = chanend_read_word(d);
            if (v) {
              while (!chanend_nomore_byte(c)) {
                chanend_write_byte(d, chanend_read_byte(c));
              }
              chanend_all_sent(d);
              core[k].mem -= v;
              core[k].thr -= 1;
              r = chanend_read_word(d);
              chanend_all_received(d);
#ifdef DEBUG_DISPATCH
              printstr(log, "*** DP CTRL ");
              printhex(log, r, 8);
              printchar(log, ' ');
              printhex(log, v, 8);
              println(log);
#endif
            } else {
#ifdef DEBUG_DISPATCH
              printstr(log, "*** DP MEM FAULT");
              println(log);
#endif
              chanend_read_strip(c);
              chanend_hangup(d);
            }
          } else {
#ifdef DEBUG_DISPATCH
            printstr(log, "*** DP OUT OF MEM");
            println(log);
#endif
            chanend_read_strip(c);
          }
        } else {
#ifdef DEBUG_DISPATCH
          printstr(log, "*** DP MAGIC FAULT");
          println(log);
#endif
          chanend_read_strip(c);
        }
        chanend_write_word(c, r);
        chanend_hangup(c);
        break;
      case dp_cmd_done:
#ifdef DEBUG_DISPATCH
        printstr(log, "*** DP DONE ");
        printhex(log, v, 8);
        println(log);
#endif
        chanend_hangup(c);
        for (i = 0; i < cores; i++) {
          if (resource_compatible(core[i].sched, v, 1, 1, 0, 0))
            break;
        }
        chanend_connect(d, core[i].sched);
        chanend_write_byte(d, sm_cmd_done);
        chanend_write_word(d, v);
        core[i].thr = chanend_read_word(d);
        core[i].mem = chanend_read_word(d);
        chanend_hangup(d);
#ifdef DEBUG_DISPATCH
        printstr(log, "*** DP STAT ");
        printhex(log, core[i].thr, 2);
        printchar(log, ' ');
        printhex(log, core[i].mem, 8);
        println(log);
#endif
        break;
      default:
        chanend_hangup(c);
        break;
    }
  } while (1);
}

void init0(void)
{
  resource t, r, log;
  unsigned s0[768];
  unsigned s1[256];
  unsigned s2[256];
  unsigned s3[256];

  t = timer_book();
  timer_wait(t, time_per_somesixthsecond);
  r = thread_book();
  thread_start_preliminary(r, &reception, &s0[767]);
  timer_wait(t, (1 << 8) - 1);
  r = thread_book();
  thread_start_preliminary(r, &dispatch, &s1[255]);
  r = thread_book();
  thread_start_preliminary(r, &uart_rx, &s2[255]);
  r = thread_book();
  thread_start_preliminary(r, &uart_tx, &s3[255]);
  log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
  timer_wait(t, (1 << 24) - 1);
  timer_free(t);
  execuser("init");
  schedule(log);
}

void init1(void)
{
  resource t, c, log;

  t = timer_book();
  timer_wait(t, time_per_somesixthsecond);
  timer_wait(t, (1 << 8) - 1);
  timer_free(t);
  log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
  schedule(log);
}

void init2(void)
{
  resource t, r, c, log;
  unsigned s1[256];
  unsigned s2[256];
  unsigned s3[256];
  unsigned s4[256];
  unsigned s5[256];
  unsigned s6[256];

  t = timer_book();
  timer_wait(t, time_per_somesixthsecond);
  timer_wait(t, (1 << 8) - 1);
  timer_free(t);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_init, &s1[255]);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_rx, &s2[255]);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_tx, &s3[255]);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_arp, &s4[255]);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_ipv4, &s5[255]);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_icmp, &s6[255]);
  log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
  schedule(log);
}

void init3(void)
{
  resource t, r, log;
  unsigned s1[256];
  unsigned s2[256];
  unsigned s3[256];

  t = timer_book();
  timer_wait(t, time_per_somesixthsecond);
  timer_wait(t, (1 << 8) - 1);
  timer_free(t);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_udp, &s1[255]);
  r = thread_book();
  thread_start_preliminary(r, &ethernet_tftp, &s2[255]);
  log = chanend_book();
  chanend_destination(log, chanend_cr_request(log, cr_uart_tx));
  schedule(log);
}
