/* copyright (c)2012 Oskar Schirmer <oskar@scara.com> */
#include "chanend.h"
#include "reception.h"
#include "resource.h"
#include "uart.h"

#define baud_rate 115200
#define time_per_bit (time_per_second / baud_rate)
#define uart_txd_pin 7
#define uart_rxd_pin 5

#ifdef TOGGLE
static void toggle(void)
{
  resource p;
  static v;
  v = v ? 0 : 1;
  p = port_book(1, 9); /* K   */
  port_on(p);
  port_write(p, v);
  p = port_book(1, 10); /* I  LED 0 */
  port_on(p);
  port_write(p, v);
}
#endif

void uart_rx(void)
{
  resource c, t, p;
  time n;
  unsigned d;
  c = chanend_book();
  chanend_cr_welcome(c, cr_uart_rx);
  t = timer_book();
  p = port_book(1, uart_rxd_pin);
  port_on(p);
  do {
    chanend_listen(c);
    event_vector(c, uart_rx_chan);
    event_vector(p, uart_rx_start);
    event_vector(t, uart_rx_sample);
    do {
      port_config_cond(p, port_cond_eq, 0);
      timer_watch(t);
      event_book();
      event_enable(c);
      event_enable(p);
      event_await();
      event_avoid_dce(uart_rx_chan);
      event_avoid_dce(uart_rx_start);
event_label(uart_rx_start):
      n = timer_read(t) + time_per_bit / 2;
      port_read(p);
      port_config_cond(p, port_cond_none, 0);
      d = 1 << 9;
      timer_alarm(t);
      event_book();
      event_enable(c);
      event_enable(t);
      do {
        n += time_per_bit;
        timer_set(t, n);
        event_await();
        event_avoid_dce(uart_rx_sample);
event_label(uart_rx_sample):
        timer_read(t);
        d = port_read_shift(p, d);
#ifdef TOGGLE
        toggle();
#endif
      } while (!(d & 1));
      if (((signed)d) < 0) {
        d = d >> 23;
        chanend_write_byte(c, d);
      }
    } while (1);
event_label(uart_rx_chan):
    chanend_hangup(c);
  } while (1);
  chanend_free(c);
}

void uart_tx(void)
{
  resource c, t, p;
  time n;
  unsigned d;
  c = chanend_book();
  chanend_cr_welcome(c, cr_uart_tx);
  t = timer_book();
  p = port_book(1, uart_txd_pin);
  port_on(p);
  port_write(p, 1);
  do {
    while (!chanend_nomore_byte(c)) {
      d = chanend_read_byte(c);
      d = (d | 0x100) << 1;
      timer_watch(t);
      n = timer_read(t);
      timer_alarm(t);
      do {
        d = port_write_shift(p, d);
        n += time_per_bit;
        timer_set(t, n);
        timer_read(t);
      } while (d);
    }
    chanend_all_received(c);
  } while (1);
  port_free(p);
  timer_free(t);
  chanend_free(c);
}
